#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Interfaces for casa_envoy
"""
import abc
import logging
import os
import subprocess
import sys
from abc import ABC


class LauncherIF(ABC):
    """
    Generic Launcher methods.
    Should be implemented for any type of CASA processing.
    """

    @abc.abstractmethod
    def launch_casa(self):
        pass

    @abc.abstractmethod
    def prepare_for_casa(self):
        pass


class RunnerABC(ABC):
    def __init__(self, parameters: dict):
        self.logger = logging.getLogger("casa_envoy")
        self.parameters = parameters

    @property
    @abc.abstractmethod
    def build_command(self) -> str:
        pass

    @property
    def log_message(self):
        return f"RUNNING CASA with command '{self.build_command}'"

    def call_casa(self) -> tuple[bytes, bytes]:
        self.logger.info(self.log_message)
        os.chdir("./working")
        result = subprocess.Popen(
            self.build_command,
            shell=True,
            executable="/bin/bash",
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
        return result.communicate()

    def __eq__(self, __value: object) -> bool:
        if not __value.__class__ == self.__class__:
            return False
        return self.parameters == __value.parameters and self.build_command == __value.build_command


class AuditorIF(ABC):
    """
    Generic functionality implementation for auditor classes
    """

    @abc.abstractmethod
    def audit(self):
        pass


class FoundationIF(ABC):
    """
    Generic Foundation methods
    Should be implemented for any type of CASA processing that requires initial data placement
    in locations other than rawdata after download
    """

    @abc.abstractmethod
    def data_foundation(self):
        pass
