#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
from pathlib import Path
from unittest.mock import mock_open, patch

import pytest
from casa_envoy.configure import ParallelRunner, PiperestorescriptRunner, StandardRunner
from casa_envoy.enums import ProductType
from casa_envoy.launchers import (
    CalibrationLauncher,
    CasaLauncher,
    ImagingLauncher,
    determine_runner_type,
    get_abs_file,
    run_audit,
)
from casa_envoy.schema import AbstractTextFile

piperestorescript_parameters = {
    "useCasa": False,
    "homeForReprocessing": "/home/casa/packages/pipeline/current",
    "rootDirectory": "/tmp/workspaces_tmp/",
    "processingDirectory": "tmpiox5trbp",
    "metadata": "test/input_files/restore.json",
    "ppr": "test/input_files/PPR.xml",
    "product_type": "restore",
    "parent_path": Path("I/am/a/path"),
    "piperestorescript_path": Path("path/to/casapiperestorescript.py"),
}
cube_parameters = {
    "useCasa": False,
    "homeForReprocessing": "/home/casa/packages/pipeline/current",
    "rootDirectory": "/tmp/workspaces_tmp/",
    "processingDirectory": "tmpiox5trbp",
    "metadata": "test/input_files/cube-metadata.json",
    "ppr": "test/input_files/cmsimage-PPR.xml",
    "product_type": "standard-img",
    "parent_path": Path("I/am/a/path"),
}

restore_parameters = {
    "useCasa": False,
    "homeForReprocessing": "/home/casa/packages/pipeline/current",
    "rootDirectory": "/tmp/workspaces_tmp/",
    "processingDirectory": "tmpiox5trbp",
    "metadata": "test/input_files/restore.json",
    "ppr": "test/input_files/PPR.xml",
    "product_type": "restore",
    "parent_path": Path("I/am/a/path"),
}


@pytest.fixture
def std_runner(cal_settings) -> StandardRunner:
    return StandardRunner(parameters=cal_settings)


@pytest.fixture
def parallel_runner(img_settings) -> ParallelRunner:
    return ParallelRunner(parameters=img_settings)


def test_get_abs_file():
    """
    Verify abstract text file creation

    :return:
    """
    with patch("builtins.open", mock_open(read_data="I am content"), create=True) as content:
        file = get_abs_file("test/input_files/restore.json")
        assert isinstance(file, AbstractTextFile)
        assert isinstance(file.filename, str)
        assert isinstance(file.content, str)


@patch("sys.exit")
@patch("logging.getLogger")
def test_run_audit(mock_logger, mock_exit, cal_settings):
    """
    Verify auditing runs

    :param mock_logger:
    :param mock_exit:
    :return:
    """
    with patch("casa_envoy.auditor.AuditDirectories.audit") as dir_audit:
        with patch("casa_envoy.auditor.AuditFiles.audit") as file_audit:
            meta_file = get_abs_file("test/input_files/test.json")
            ppr_file = get_abs_file("test/input_files/PPR.xml")
            run_audit(mock_logger, ppr_file, meta_file, cal_settings)
            assert mock_logger.call_count == 2
            assert file_audit.call_count == 1
            assert dir_audit.call_count == 1


def test_determine_runner_type(cal_settings):
    """
    Verify correct typed runner is created

    :return:
    """
    runner = determine_runner_type(parameters=cal_settings)
    assert isinstance(runner, StandardRunner)

    try:
        cal_settings["requested_parallel"] = 4
        runner = determine_runner_type(parameters=cal_settings)
        assert isinstance(runner, ParallelRunner)
    finally:
        # reset
        del cal_settings["requested_parallel"]


class TestCasaLauncher:
    """
    Test methods for CASA launcher setup
    """

    @patch("subprocess.run")
    @patch("os.chdir")
    def test_run(self, mock_os, mock_subprocess, std_runner, parallel_runner):
        with patch("casa_envoy.launchers.CasaLauncher.check_logs") as logs:
            with patch("casa_envoy.configure.StandardRunner.call_casa") as casa:
                CasaLauncher(runner=std_runner).run()
                assert mock_subprocess.call_count == 1
            with patch("casa_envoy.configure.ParallelRunner.call_casa") as casa:
                CasaLauncher(runner=parallel_runner).run()
                assert mock_subprocess.call_count == 2

    @patch("builtins.open")
    @patch("glob.glob")
    @patch("os.chdir")
    @patch("os.getcwd")
    def test_check_logs(self, mock_os_cwd, mock_os_dir, mock_glob, mock_o, std_runner):
        CasaLauncher(runner=std_runner).check_logs()
        assert mock_os_cwd.call_count == 1
        assert mock_os_dir.call_count == 0
        assert mock_glob.call_count == 1


class TestCalibrationLauncher:
    """
    Test methods for calibration launcher setup
    """

    @patch("casa_envoy.launchers.run_audit")
    @patch("casa_envoy.launchers.CasaLauncher.run")
    def test_launch_casa(self, mock_run, mock_audit, cal_settings):
        CalibrationLauncher(parameters=cal_settings).launch_casa()
        assert mock_run.call_count == 1
        assert mock_audit.call_count == 1

    def test_cal_prepare_for_casa(self, cal_settings):
        test_cases = [
            {"parameters": cal_settings, "restore_call_count": 0, "runner": StandardRunner(cal_settings)},
            {"parameters": restore_parameters, "restore_call_count": 1, "runner": StandardRunner(restore_parameters)},
            {
                "parameters": piperestorescript_parameters,
                "restore_call_count": 1,
                "runner": PiperestorescriptRunner(piperestorescript_parameters),
            },
        ]
        for test_case in test_cases:
            with patch("casa_envoy.foundation.GeneralFoundation.data_foundation") as general:
                with patch("casa_envoy.foundation.RestoreFoundation.data_foundation") as restore:
                    launcher = CalibrationLauncher(parameters=test_case["parameters"])
                    launcher.prepare_for_casa()
                    assert general.call_count == 1
                    assert restore.call_count == test_case["restore_call_count"]
                    assert launcher.runner == test_case["runner"]

    def test_check_calibratable(self, cal_settings):
        check = CalibrationLauncher(parameters=cal_settings).check_calibratable()
        assert check is True

    def test_check_restorable(self, cal_settings):
        test_cases = [
            {"parameters": restore_parameters, "is_restorable": True},
            {"parameters": piperestorescript_parameters, "is_restorable": True},
            {"parameters": cal_settings, "is_restorable": False},
        ]
        for test_case in test_cases:
            check = CalibrationLauncher(parameters=test_case["parameters"]).check_restorable()
            assert check is test_case["is_restorable"]


class TestImagingLauncher:
    """
    Tests methods for imaging launcher setup
    """

    @patch("casa_envoy.launchers.run_audit")
    @patch("casa_envoy.launchers.CasaLauncher.run")
    def test_launch_casa(self, mock_run, mock_audit, img_settings):
        CalibrationLauncher(parameters=img_settings).launch_casa()
        assert mock_run.call_count == 1
        assert mock_audit.call_count == 1

    def test_img_prepare_for_casa(self, img_settings):
        with patch("casa_envoy.foundation.GeneralFoundation.data_foundation") as general:
            with patch("casa_envoy.foundation.RestoreFoundation.data_foundation") as restore:
                with patch("casa_envoy.foundation.CubeFoundation.data_foundation") as cube:
                    ImagingLauncher(parameters=img_settings).prepare_for_casa()
                    assert general.call_count == 1
                    assert restore.call_count == 0
                    assert cube.call_count == 0
                    # reset
                    general.call_count = 0

                    ImagingLauncher(parameters=restore_parameters).prepare_for_casa()
                    assert general.call_count == 1
                    assert restore.call_count == 1
                    assert cube.call_count == 0
                    # reset
                    general.call_count = 0
                    restore.call_count = 0

                    ImagingLauncher(parameters=cube_parameters).prepare_for_casa()
                    assert general.call_count == 1
                    assert restore.call_count == 0
                    assert cube.call_count == 1

    def test_check_imageable(self, img_settings):
        check = ImagingLauncher(parameters=img_settings).check_imageable()
        assert check is True

    def test_check_cal_and_imageable(self):
        restore_parameters["product_type"] = ProductType.INTEGRATED_CAL_IMG.value
        check = ImagingLauncher(parameters=restore_parameters).check_cal_and_imageable()
        assert check is True
